// Multiple-reader/Single-writer Lock in Windows NT
// Author: Jeff Claar

#ifndef __RWLOCK_HPP__
#define __RWLOCK_HPP__

class SyRWLock
{
private:
    // RTL_RWLOCK structure.  This is the
    // structure used by ntdll.dll to handle
    // the MRSW resource.
    struct RTL_RWLOCK
    {
        CRITICAL_SECTION    mCS;
        HANDLE              mSem1;
        DWORD               mSharedWaiters;
        HANDLE              mSem2;
        DWORD               mExclusiveWaiters;
        DWORD               mNumberActive;
        DWORD               mOwningThreadId;
        DWORD               mU2;
        PVOID               mDebugInfo;
    };

    typedef RTL_RWLOCK*     PRTL_RWLOCK;

    RTL_RWLOCK              mRWLock;

    // Function prototypes for MRSW resources.
    typedef VOID
        (WINAPI*RtlInitializeResource)(PRTL_RWLOCK);
    typedef VOID
        (WINAPI *RtlDeleteResource)(PRTL_RWLOCK);
    typedef BYTE
        (WINAPI *RtlAcqResourceExclusive)(PRTL_RWLOCK, BYTE);
    typedef BYTE
        (WINAPI *RtlAcqResourceShared)(PRTL_RWLOCK, BYTE);
    typedef VOID
        (WINAPI *RtlReleaseResource)(PRTL_RWLOCK);
    typedef VOID
        (WINAPI *RtlDumpResource)(PRTL_RWLOCK);

    // Names for functions as exported from ntdll.dll.
    static LPCSTR sInitName;
    static LPCSTR sDeleteName;
    static LPCSTR sGetExclusiveName;
    static LPCSTR sGetSharedName;
    static LPCSTR sReleaseName;
    static LPCSTR sDumpName;

    // Static members containing function pointers.
    static RtlInitializeResource sInitProc;
    static RtlDeleteResource sDelProc;
    static RtlAcqResourceExclusive sGetExclusiveProc;
    static RtlAcqResourceShared sGetSharedProc;
    static RtlReleaseResource sReleaseProc;
    static RtlDumpResource sDumpProc;

public:
    SyRWLock();
    ~SyRWLock();

    // Public members to lock and
    // release the MRSW resource.
    BYTE GetExclusive(BYTE fWait);
    BYTE GetShared(BYTE fWait);
    void Release();

    // Function to dump the current status of the
    // resource to the debugger output.
    void Dump();
};

#endif
