// THRDMON.cpp - main module for THRDMON.VxD.
// ===========   
// Implements a simple Thread Monitor.Thread Monitor to 
// demonstrates the usage of Windows 95 LOGGER.VXD services. 
// Use \bin\loader.exe to load/unload this VxD.
//
// 04/01/98 VV
//
#define DEVICE_MAIN
#include "thrdmon.h"
Declare_Virtual_Device(THRDMON)
#undef DEVICE_MAIN

#include    <logR0api.h> // LOGGER.VxD Service API

// log file handle
static HANDLE  hLogFile;

// VxD is being loaded: try to open the log file
BOOL ThrdmonDevice::OnSysDynamicDeviceInit()
{
    DWORD ver = LOGR0_GetVersion();
    if (ver == 0)
        return FALSE;   // problem: logger.vxd is not available
    hLogFile = LOGR0_OpenFile("ThreadMonLog.txt", NULL);
    if (hLogFile == NULL)
        return FALSE;   // problem: can't open the log file
    LOGR0_Printf(hLogFile, "Opened ThreadMon Log\n");
    return TRUE;
}

// VxD is being unloaded: close the log file
BOOL ThrdmonDevice::OnSysDynamicDeviceExit()
{
    if (hLogFile) {
        LOGR0_Printf(hLogFile, "Closed ThreadMon Log\n");
        LOGR0_CloseFile(hLogFile);
    }

    return TRUE;
}

// create an instance to hook to the thread events and log the fact:
BOOL ThrdmonDevice::OnCreateThread(THREADHANDLE hThread)
{
    new ThrdmonThread(hThread);
    LOGR0_Printf(hLogFile, 
        "Thread %X got CREATED; %d threads are being monitored\n",
            hThread, ThrdmonThread::TotalCount);
    return TRUE;
}        

// Total count of threads being monitored
DWORD ThrdmonThread::TotalCount = 0;

// thread hook constructor
ThrdmonThread::ThrdmonThread(THREADHANDLE hThread):VThread(hThread)
{
    TotalCount++;
}

// destroy the instance and log the fact
VOID ThrdmonThread::OnDestroyThread()
{
    LOGR0_Printf(hLogFile, 
        "Thread %X got DESTROYED; it ran for %d msec\n",
            m_handle, GetThreadExecTime(m_handle));
    TotalCount--;

    // Since this is the last event for this object, we have to:
    delete this;
}

// end of file
