// wdjsrvc.c

#include <windows.h>

HANDLE  hDoneEvent = NULL, hThread = NULL;
DWORD   dwCurrentState;
SERVICE_STATUS_HANDLE  hService;

BOOL NotifySCM(DWORD, DWORD, DWORD);
VOID WINAPI ServiceMain(DWORD, LPTSTR *);
VOID WINAPI ServiceHandler(DWORD);
DWORD WINAPI WorkerThread(LPVOID);

//-----------------------------------------------------------------
void main(int argc, char **argv)
{
    SERVICE_TABLE_ENTRY ServiceTable[] = {
        {TEXT("WDJSrvc"), ServiceMain}, {NULL, NULL}
    };
    StartServiceCtrlDispatcher(ServiceTable);
}

//-----------------------------------------------------------------
VOID WINAPI ServiceMain(DWORD dwArgc, LPTSTR *lpszArgv)
{
    DWORD   ThreadId;

    if (!(hService = RegisterServiceCtrlHandler(
                           TEXT("WDJSrvc"), ServiceHandler))) {
        return;
    }

    NotifySCM(SERVICE_START_PENDING, 0 , 1);

    hDoneEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    hThread = CreateThread(0,0,WorkerThread,0,0,&ThreadId);

    NotifySCM(SERVICE_RUNNING, 0 , 0);
    WaitForSingleObject(hDoneEvent, INFINITE);

    CloseHandle(hThread);
    CloseHandle(hDoneEvent);
    return;
}

//-----------------------------------------------------------------
VOID WINAPI ServiceHandler(DWORD fdwControl)
{
    switch(fdwControl) {
    case SERVICE_CONTROL_STOP:
        NotifySCM(SERVICE_STOP_PENDING, 0 , 1);
        SetEvent(hDoneEvent);
        NotifySCM(SERVICE_STOPPED, 0 , 0);
        break;

    case SERVICE_CONTROL_INTERROGATE:
        NotifySCM(dwCurrentState, 0 , 0);
        break;
   
    case SERVICE_CONTROL_SHUTDOWN:
        break;
    }
}

//-----------------------------------------------------------------
BOOL NotifySCM(DWORD State, DWORD ExitCode, DWORD Progress)
{
    SERVICE_STATUS ServiceStatus;

    ServiceStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
    ServiceStatus.dwCurrentState = dwCurrentState = State;
    ServiceStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP |
                                       SERVICE_ACCEPT_SHUTDOWN;
    ServiceStatus.dwWin32ExitCode = ExitCode;
    ServiceStatus.dwServiceSpecificExitCode = 0;
    ServiceStatus.dwCheckPoint = Progress;
    ServiceStatus.dwWaitHint = 5000;

    return SetServiceStatus(hService, &ServiceStatus);
}

//-----------------------------------------------------------------
DWORD WINAPI WorkerThread(LPDWORD ThreadParam)
{
   HANDLE hEvent, hEventLog;
   EVENTLOGRECORD *pLog;  
   BYTE Buffer[1024]; 
   DWORD Records,Oldest,dwBytesRead, dwBytesNext; 
   TCHAR szMsg[128];

   hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
   
   // priming read - position evenlog seek position at the end

   hEventLog = OpenEventLog(NULL, TEXT("Security"));
   GetNumberOfEventLogRecords(hEventLog, &Records);
   GetOldestEventLogRecord(hEventLog, &Oldest);
   pLog = (EVENTLOGRECORD *)&Buffer; 
   ReadEventLog(hEventLog, EVENTLOG_SEEK_READ |
                EVENTLOG_FORWARDS_READ, Records+Oldest-1, pLog, 
                1024, &dwBytesRead, &dwBytesNext);

   NotifyChangeEventLog(hEventLog, hEvent);
 
   while (TRUE) {                                   
      WaitForSingleObject(hEvent, INFINITE);              
      
      // read the first new event log record

      pLog = (EVENTLOGRECORD *)&Buffer; 
      if (!ReadEventLog(hEventLog,
                   EVENTLOG_FORWARDS_READ|EVENTLOG_SEQUENTIAL_READ,
                   0, pLog, 1024, &dwBytesRead, &dwBytesNext)) {
          
         if (GetLastError() == ERROR_EVENTLOG_FILE_CHANGED) {            
            
            // probably the eventlog was cleared, get fresh handle

            CloseEventLog(hEventLog);
            hEventLog = OpenEventLog(NULL, TEXT("Security"));
            GetNumberOfEventLogRecords(hEventLog, &Records);
            GetOldestEventLogRecord(hEventLog, &Oldest);
            pLog = (EVENTLOGRECORD *)&Buffer; 
              
            ReadEventLog(hEventLog, EVENTLOG_SEEK_READ |
                         EVENTLOG_FORWARDS_READ, Records+Oldest-1, pLog, 
                         1024, &dwBytesRead, &dwBytesNext);
            
            NotifyChangeEventLog(hEventLog, hEvent);
            continue;
         }
      }

      do { 
         GetNumberOfEventLogRecords(hEventLog, &Records);
         GetOldestEventLogRecord(hEventLog, &Oldest);
           
         while (dwBytesRead > 0) { 
            wsprintf(szMsg, TEXT("New Event: %d,%d,%d"),
                     pLog->EventID,Records,Oldest);
            MessageBox(NULL, szMsg, TEXT("WDJSrvc"),
                       MB_OK | MB_SERVICE_NOTIFICATION);
            dwBytesRead -= pLog->Length; 
            pLog = (EVENTLOGRECORD *)((LPBYTE)pLog + pLog->Length); 
         } 
         pLog = (EVENTLOGRECORD *)&Buffer; 
      } while (ReadEventLog(hEventLog,
                   EVENTLOG_FORWARDS_READ|EVENTLOG_SEQUENTIAL_READ,
                   0, pLog, 1024, &dwBytesRead, &dwBytesNext));
   }
   return TRUE;
}

